@@ -14,36 +14,58 @@
 #include <minizinc/model.hh>
 #include <minizinc/prettyprinter.hh>
 
+#include <algorithm>
 #include <unordered_set>
+#include <vector>
 
 #undef MZN_DEBUG_FUNCTION_REGISTRY
 
 namespace MiniZinc {
 
-Model::FnEntry::FnEntry(FunctionI* fi0)
+Model::FnEntry::FnEntry(EnvI& env, FunctionI* fi0)
     : t(fi0->paramCount()), fi(fi0), isPolymorphic(false), isPolymorphicVariant(false) {
   for (unsigned int i = 0; i < fi->paramCount(); i++) {
     t[i] = fi->param(i)->type();
-    isPolymorphic |= (t[i].bt() == Type::BT_TOP);
+    if (t[i].bt() == Type::BT_TUPLE && t[i].typeId() == 0) {
+      env.registerTupleType(fi->param(i)->ti());
+      t[i] = fi->param(i)->ti()->type();
+      fi->param(i)->type(t[i]);
+    } else if (t[i].bt() == Type::BT_RECORD && t[i].typeId() == 0) {
+      env.registerRecordType(fi->param(i)->ti());
+      t[i] = fi->param(i)->ti()->type();
+      fi->param(i)->type(t[i]);
+    }
+    isPolymorphic |= checkPoly(env, t[i]);
   }
 }
 
-bool Model::FnEntry::operator<(const Model::FnEntry& f) const {
-  assert(!compare(*this, f) || !compare(f, *this));
-  return compare(*this, f);
+bool Model::FnEntry::checkPoly(const EnvI& env, const Type& t) {
+  if (t.bt() == Type::BT_TOP) {
+    return true;
+  }
+  if (t.structBT()) {
+    StructType* st = env.getStructType(t);
+    for (size_t i = 0; i < st->size(); ++i) {
+      if (checkPoly(env, (*st)[i])) {
+        return true;
+      }
+    }
+  }
+  return false;
 }
 
-bool Model::FnEntry::compare(const Model::FnEntry& e1, const Model::FnEntry& e2) {
+bool Model::FnEntry::compare(const EnvI& env, const Model::FnEntry& e1, const Model::FnEntry& e2) {
   if (e1.t.size() < e2.t.size()) {
     return true;
   }
   if (e1.t.size() == e2.t.size()) {
     for (unsigned int i = 0; i < e1.t.size(); i++) {
       if (e1.t[i] != e2.t[i]) {
-        if (e1.t[i].isSubtypeOf(e2.t[i], true)) {
+        if (e1.t[i].isSubtypeOf(env, e2.t[i], true)) {
+          assert(!e2.t[i].isSubtypeOf(env, e1.t[i], true));
           return true;
         }
-        if (e2.t[i].isSubtypeOf(e1.t[i], true)) {
+        if (e2.t[i].isSubtypeOf(env, e1.t[i], true)) {
           return false;
         }
         switch (e1.t[i].cmp(e2.t[i])) {
@@ -118,6 +140,24 @@ void Model::setOutputItem(OutputI* oi) {
 namespace {
 
 enum PossibleBaseTypes { PBT_ANY, PBT_ALL, PBT_BIFS, PBT_BIF, PBT_I };
+PossibleBaseTypes pbt_from_type(const Type& t) {
+  if (t.any()) {
+    return PBT_ANY;
+  }
+  if (t.isvar()) {
+    if (t.isOpt()) {
+      return PBT_BIF;
+    }
+    if (t.isSet()) {
+      return PBT_I;
+    }
+    return PBT_BIFS;
+  }
+  if (t.isSet()) {
+    return PBT_BIF;
+  }
+  return PBT_ALL;
+}
 PossibleBaseTypes pbt_join(PossibleBaseTypes pbt0, PossibleBaseTypes pbt1) {
   if (pbt0 == PBT_ANY) {
     return pbt1;
@@ -167,6 +207,7 @@ bool can_increment_type(const Type& t, PossibleBaseTypes pbt) {
     case PBT_I:
       return false;
   }
+  throw InternalError("Invalid BaseType");
 }
 void increment_type(Type& t, PossibleBaseTypes pbt) {
   assert(pbt != PBT_I);
@@ -246,9 +287,72 @@ void increment_type(Type& t, PossibleBaseTypes pbt) {
 
 }  // namespace
 
-void Model::addPolymorphicInstances(Model::FnEntry& fe, std::vector<FnEntry>& entries) {
-  entries.push_back(fe);
+// For each TIId, store pointers to the type objects that refer to this TIId
+struct TIIDInfo {
+  std::vector<Type*> t;
+  PossibleBaseTypes pbt;
+  TIIDInfo(std::vector<Type*> t0, PossibleBaseTypes pbt0) : t(std::move(t0)), pbt(pbt0) {}
+};
+
+void TypeInst::collectTypeIds(std::unordered_map<ASTString, size_t>& seen_tiids,
+                              std::vector<TIIDInfo>& type_ids) const {
+  auto* al = domain()->cast<ArrayLit>();
+  for (size_t i = 0; i < al->size(); i++) {
+    auto* ti = (*al)[i]->cast<TypeInst>();
+    if (ti->type().bt() == Type::BT_TOP) {
+      assert(ti->domain() && ti->domain()->isa<TIId>());
+      TIId* id0 = ti->domain()->cast<TIId>();
+      auto it = seen_tiids.find(id0->v());
+      if (it == seen_tiids.end()) {
+        type_ids.emplace_back(
+            std::vector<Type*>({&ti->_type}),
+            pbt_join(ti->type().any() ? PBT_ANY : PBT_ALL, pbt_from_type(ti->type())));
+        seen_tiids.emplace(id0->v(), type_ids.size() - 1);
+      } else {
+        TIIDInfo& info = type_ids[it->second];
+        info.pbt = pbt_join(info.pbt, pbt_from_type(ti->type()));
+        info.t.push_back(&ti->_type);
+      }
+    } else if (ti->type().structBT()) {
+      size_t size = seen_tiids.size();
+      ti->collectTypeIds(seen_tiids, type_ids);
+      if (size < seen_tiids.size()) {
+        ti->_type.typeId(0);
+      }
+    }
+  }
+}
+
+void Model::addPolymorphicInstances(EnvI& env, Model::FnEntry& fe, std::vector<FnEntry>& entries) {
+  auto addEntry = [&](Model::FnEntry& toAdd) {
+    for (auto& entry : entries) {
+      if (entry.t == toAdd.t) {
+        bool more_specific = true;
+        for (unsigned int i = 0; i < toAdd.fi->paramCount(); i++) {
+          // If all parameters of the entry we are adding are subtypes of an
+          // existing function, this one should take priority
+          //
+          // E.g. (bool, int) is preferred over ($T, $U)
+          auto oldParamType = entry.fi->param(i)->type();
+          auto newParamType = toAdd.fi->param(i)->type();
+          if (!newParamType.isSubtypeOf(env, oldParamType, false)) {
+            more_specific = false;
+            break;
+          }
+        }
+        if (more_specific) {
+          entry = toAdd;
+        }
+        return;
+      }
+    }
+    // Entry not yet added
+    entries.push_back(toAdd);
+  };
+
+  addEntry(fe);
   if (fe.isPolymorphic) {
+    GCLock lock;
     FnEntry cur = fe;
     cur.isPolymorphicVariant = true;
 
@@ -271,57 +375,13 @@ void Model::addPolymorphicInstances(Model::FnEntry& fe, std::vector<FnEntry>& en
 
      */
 
-    // For each TIId, store pointers to the type objects that refer to this TIId
-    struct TIIDInfo {
-      std::vector<Type*> t;
-      PossibleBaseTypes pbt;
-      TIIDInfo(std::vector<Type*> t0, PossibleBaseTypes pbt0) : t(std::move(t0)), pbt(pbt0) {}
-    };
+    std::unordered_map<ASTString, size_t> type_id_map;
     std::vector<TIIDInfo> type_ids;
-    std::unordered_set<ASTString> seen_tiids;
 
-    // First step: initialise all type variables to bool
-    // and collect them in the type_ids vector
-    for (unsigned int i = 0; i < cur.t.size(); i++) {
-      if (cur.t[i].bt() == Type::BT_TOP) {
-        assert(cur.fi->param(i)->ti()->domain() && cur.fi->param(i)->ti()->domain()->isa<TIId>());
-        TIId* id0 = cur.fi->param(i)->ti()->domain()->cast<TIId>();
-        if (seen_tiids.find(id0->v()) == seen_tiids.end()) {
-          seen_tiids.insert(id0->v());
-          // New tiid, collect all occurrences
-          std::vector<Type*> t;
-          PossibleBaseTypes pbt = cur.fi->param(i)->ti()->type().any() ? PBT_ANY : PBT_ALL;
-          for (unsigned int j = i; j < cur.t.size(); j++) {
-            if ((cur.fi->param(j)->ti()->domain() != nullptr) &&
-                cur.fi->param(j)->ti()->domain()->isa<TIId>()) {
-              TIId* id1 = cur.fi->param(j)->ti()->domain()->cast<TIId>();
-              if (id0->v() == id1->v()) {
-                // Found parameter with same type variable
-                if (cur.fi->param(j)->ti()->type().any()) {
-                  pbt = pbt_join(pbt, PBT_ANY);
-                } else if (cur.fi->param(j)->type().isvar()) {
-                  if (cur.fi->param(j)->type().isOpt()) {
-                    pbt = pbt_join(pbt, PBT_BIF);
-                  } else if (cur.fi->param(j)->type().isSet()) {
-                    pbt = pbt_join(pbt, PBT_I);
-                  } else {
-                    pbt = pbt_join(pbt, PBT_BIFS);
-                  }
-                } else {
-                  if (cur.fi->param(j)->type().isSet()) {
-                    pbt = pbt_join(pbt, PBT_BIF);
-                  } else {
-                    pbt = pbt_join(pbt, PBT_ALL);
-                  }
-                }
-                t.push_back(&cur.t[j]);
-              }
-            }
-          }
-          type_ids.emplace_back(t, pbt);
-        }
-      }
-    }
+    // Create a parameter TypeInst in the format of a tuple TypeInst
+    // Immediately copy so the internal TypeInst values can be changed
+    auto* paramtuple = copy(env, cur.fi->paramTypes())->cast<TypeInst>();
+    paramtuple->collectTypeIds(type_id_map, type_ids);
 
     std::vector<size_t> stack;
     for (size_t i = 0; i < type_ids.size(); i++) {
@@ -334,17 +394,20 @@ void Model::addPolymorphicInstances(Model::FnEntry& fe, std::vector<FnEntry>& en
 
     while (!stack.empty()) {
       if (stack.back() == final_id) {
-        // If this instance isn't in entries yet, add it
-        bool alreadyDefined = false;
-        for (auto& entry : entries) {
-          if (entry.t == cur.t) {
-            alreadyDefined = true;
-            break;
+        // New complete instance
+        // First, update cur types
+        auto* tis = paramtuple->domain()->cast<ArrayLit>();
+        for (size_t i = 0; i < tis->size(); ++i) {
+          cur.t[i] = (*tis)[i]->type();
+          if (cur.t[i].bt() == Type::BT_TUPLE && cur.t[i].typeId() == 0) {
+            env.registerTupleType((*tis)[i]->cast<TypeInst>());
+            cur.t[i] = (*tis)[i]->type();
+          } else if (cur.t[i].bt() == Type::BT_RECORD && cur.t[i].typeId() == 0) {
+            env.registerRecordType((*tis)[i]->cast<TypeInst>());
+            cur.t[i] = (*tis)[i]->type();
           }
         }
-        if (!alreadyDefined) {
-          entries.push_back(cur);
-        }
+        addEntry(cur);
       }
 
       const Type& back_t = *type_ids[stack.back()].t[0];
@@ -376,8 +439,8 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
   if (i_id == m->_fnmap.end()) {
     // new element
     std::vector<FnEntry> v;
-    FnEntry fe(fi);
-    addPolymorphicInstances(fe, v);
+    FnEntry fe(env, fi);
+    addPolymorphicInstances(env, fe, v);
     m->_fnmap.insert(std::pair<ASTString, std::vector<FnEntry> >(fi->id(), v));
   } else {
     // add to list of existing elements
@@ -392,13 +455,11 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
         for (unsigned int j = 0; j < fi->paramCount(); j++) {
           Type t1 = i.fi->param(j)->type();
           Type t2 = fi->param(j)->type();
-          t1.enumId(0);
-          t2.enumId(0);
           if (t1 != t2) {
             alleq = false;
           }
-          t1.ti(Type::TI_PAR);
-          t2.ti(Type::TI_PAR);
+          t1.mkPar(env);
+          t2.mkPar(env);
           if (t1 != t2) {
             eqExceptInst = false;
           }
@@ -416,7 +477,17 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
             if (Call* deprecated = i.fi->ann().getCall(env.constants.ann.mzn_deprecated)) {
               fi->ann().add(deprecated);
             }
-            i = fi;
+            FunctionI* old_fi = i.fi;
+            i = FnEntry(env, fi);
+            // If we are replacing a polymorphic function using a new polymorphic function, then
+            // replace in all entries generated using addPolymorphicInstances
+            if (i.isPolymorphic) {
+              for (auto& j : v) {
+                if (j.fi == old_fi) {
+                  j.fi = fi;
+                }
+              }
+            }
           } else if (Call* deprecated = fi->ann().getCall(env.constants.ann.mzn_deprecated)) {
             i.fi->ann().add(deprecated);
           }
@@ -425,12 +496,16 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
         if (eqExceptInst) {
           Type t1 = i.fi->ti()->type();
           Type t2 = fi->ti()->type();
-          t1.enumId(0);
-          t2.enumId(0);
-          t1.ti(Type::TI_PAR);
-          t2.ti(Type::TI_PAR);
-          t1.ot(Type::OT_PRESENT);
-          t2.ot(Type::OT_PRESENT);
+          t1.mkPar(env);
+          t2.mkPar(env);
+          t1.mkPresent(env);
+          t2.mkPresent(env);
+          if (t1.bt() == Type::BT_INT) {
+            t1.typeId(0);
+          }
+          if (t2.bt() == Type::BT_INT) {
+            t2.typeId(0);
+          }
           if (t1 != t2) {
             throw TypeError(env, fi->loc(),
                             "function with same type up to par/var but different return type "
@@ -440,10 +515,12 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
         }
       }
     }
-    FnEntry fe(fi);
-    addPolymorphicInstances(fe, v);
+    FnEntry fe(env, fi);
+    addPolymorphicInstances(env, fe, v);
     if (keepSorted) {
-      std::sort(i_id->second.begin(), i_id->second.end());
+      std::sort(v.begin(), v.end(), [&env](const Model::FnEntry& e1, const Model::FnEntry& e2) {
+        return Model::FnEntry::compare(env, e1, e2);
+      });
     }
   }
   if (fi->id() == env.constants.ids.mzn_reverse_map_var) {
@@ -474,15 +551,17 @@ void compute_possible_matches(EnvI& env, const Model* m, const ASTString& ident,
                               std::vector<FunctionI*>& ret) {
   // Go through the types in this order: var opt, var, par opt, par
   std::vector<Type> vp(ta.size());
-  for (unsigned int i = 0; i < ta.size(); i++) {
-    vp[i] = ta[i];
-    vp[i].ti(Type::TI_VAR);
-    if (vp[i].st() == Type::ST_PLAIN) {
-      vp[i].ot(Type::OT_OPTIONAL);
+  auto reset_types_from = [&](size_t from) {
+    for (size_t i = from; i < ta.size(); i++) {
+      vp[i] = ta[i];
+      vp[i].mkVar(env);
+      if (vp[i].st() == Type::ST_PLAIN) {
+        vp[i].mkOpt(env);
+      }
     }
-  }
+  };
+  reset_types_from(0);
   int finalType = static_cast<int>(ta.size()) - 1;
-  std::vector<int> stack;
 
   for (;;) {
     auto* fi = m->matchFn(env, ident, vp, false);
@@ -495,29 +574,8 @@ void compute_possible_matches(EnvI& env, const Model* m, const ASTString& ident,
     int i = finalType;
     for (; i >= 0; i--) {
       Type& t = vp[i];
-      if (t.ti() != Type::TI_PAR || t.ot() != Type::OT_PRESENT) {
-        if (t.ti() == Type::TI_VAR) {
-          if (t.ot() == Type::OT_OPTIONAL) {
-            // var opt, turn into var
-            t.ot(Type::OT_PRESENT);
-          } else {
-            // var, turn into par opt
-            if (t.st() == Type::ST_PLAIN) {
-              t.ot(Type::OT_OPTIONAL);
-            }
-            t.ti(Type::TI_PAR);
-          }
-        } else {
-          // this is par opt, turn into par
-          t.ot(Type::OT_PRESENT);
-        }
-        for (unsigned int j = i + 1; j < ta.size(); j++) {
-          vp[j] = ta[j];
-          vp[j].ti(Type::TI_VAR);
-          if (vp[j].st() == Type::ST_PLAIN) {
-            vp[j].ot(Type::OT_OPTIONAL);
-          }
-        }
+      if (t.decrement(env)) {
+        reset_types_from(i + 1);
         break;
       }
     }
@@ -612,17 +670,20 @@ void Model::mergeStdLib(EnvI& env, Model* m) const {
       }
     }
   }
-  m->sortFn();
+  m->sortFn(env);
 }
 
-void Model::sortFn() {
+void Model::sortFn(const EnvI& env) {
   Model* m = this;
   while (m->_parent != nullptr) {
     m = m->_parent;
   }
   for (auto& it : m->_fnmap) {
     // Sort all functions by type
-    std::sort(it.second.begin(), it.second.end());
+    std::sort(it.second.begin(), it.second.end(),
+              [&env](const Model::FnEntry& e1, const Model::FnEntry& e2) {
+                return Model::FnEntry::compare(env, e1, e2);
+              });
   }
 }
 
@@ -634,7 +695,7 @@ void Model::fixFnMap() {
   for (auto& it : m->_fnmap) {
     for (auto& i : it.second) {
       for (unsigned int j = 0; j < i.t.size(); j++) {
-        if (i.t[j].isunknown()) {
+        if (i.t[j].isunknown() || i.t[j].structBT()) {
           i.t[j] = i.fi->param(j)->type();
         }
       }
@@ -642,6 +703,64 @@ void Model::fixFnMap() {
   }
 }
 
+void Model::checkFnValid(EnvI& env, std::vector<TypeError>& errors) {
+  Model* m = this;
+  while (m->_parent != nullptr) {
+    m = m->_parent;
+  }
+  // Walk through registered functions check that functions that contain a non-FlatZinc type
+  // argument are either a builtin (registed with a C++ function), or have a body.
+  for (const auto& it : m->_fnmap) {
+    const std::vector<FnEntry>& variants = it.second;
+    for (const auto& fe : variants) {
+      FunctionI* fi = fe.fi;
+      Type ret = fi->ti()->type();
+      if (fi->e() != nullptr || ret.isAnn() || fi->builtins.e != nullptr ||
+          fi->builtins.b != nullptr || fi->builtins.f != nullptr || fi->builtins.i != nullptr ||
+          fi->builtins.s != nullptr || fi->builtins.fs != nullptr || fi->builtins.str != nullptr ||
+          fi->ann().contains(env.constants.ann.mzn_internal_representation)) {
+        continue;
+      }
+      if (fi->ann().contains(env.constants.ann.output_only)) {
+        std::vector<Type> tys(fi->paramCount());
+        for (int i = 0; i < fi->paramCount(); ++i) {
+          tys[i] = fi->param(i)->type();
+          tys[i].mkPar(env);
+        }
+        fi = matchFn(env, fi->id(), tys, true);
+        ret.mkPar(env);
+        if (fi != nullptr && ret == fi->ti()->type()) {
+          continue;
+        }
+        errors.emplace_back(env, fe.fi->loc(),
+                            "Missing parameter type variant of output only function");
+        continue;
+      }
+      // Par functions that are not implemented in the compiler should have an implementation
+      if (ret.isPar()) {
+        errors.emplace_back(env, fi->loc(),
+                            "Parameter type function is missing its implementation");
+        continue;
+      }
+      if (!ret.isvarbool()) {
+        errors.emplace_back(
+            env, fi->loc(),
+            "FlatZinc builtin functions must be predicates (i.e., have `var bool` return type)");
+        continue;
+      }
+      for (int i = 0; i < fi->paramCount(); ++i) {
+        const Type& t = fi->param(i)->type();
+        if (t.isOpt() || t.structBT() || t.bt() == Type::BT_TOP) {
+          errors.emplace_back(
+              env, fi->param(i)->loc(),
+              "FlatZinc builtins are not allowed to have arguments of type " + t.toString(env));
+          break;  // Break from parameter, but does continue in FnMap
+        }
+      }
+    }
+  }
+}
+
 void Model::checkFnOverloading(EnvI& env) {
   Model* m = this;
   while (m->_parent != nullptr) {
@@ -660,8 +779,12 @@ void Model::checkFnOverloading(EnvI& env) {
         for (unsigned int i = 0; i < cur->paramCount(); i++) {
           Type t1 = cur->param(i)->type();
           Type t2 = cmp->param(i)->type();
-          t1.enumId(0);
-          t2.enumId(0);
+          if (t1.bt() == Type::BT_INT) {
+            t1.typeId(0);
+          }
+          if (t2.bt() == Type::BT_INT) {
+            t2.typeId(0);
+          }
           if (t1 != t2) {
             allEqual = false;
             break;
@@ -739,7 +862,7 @@ FunctionI* Model::matchFn(EnvI& env, const ASTString& id, const std::vector<Expr
     return matched[0];
   }
   Type t = matched[0]->ti()->type();
-  t.ti(Type::TI_PAR);
+  t.mkPar(env);
   for (unsigned int i = 1; i < matched.size(); i++) {
     if (!env.isSubtype(t, matched[i]->ti()->type(), strictEnums)) {
       throw TypeError(env, botarg->loc(), "ambiguous overloading on return type of function");
@@ -858,7 +981,7 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
     return matched[0];
   }
   Type t = matched[0]->ti()->type();
-  t.ti(Type::TI_PAR);
+  t.mkPar(env);
   for (unsigned int i = 1; i < matched.size(); i++) {
     if (!env.isSubtype(t, matched[i]->ti()->type(), strictEnums)) {
       throw TypeError(env, botarg->loc(), "ambiguous overloading on return type of function");
@@ -936,7 +1059,9 @@ FunctionI* Model::matchRevMap(EnvI& env, const Type& t0) const {
     m = m->_parent;
   }
   Type t = t0;
-  t.enumId(0);
+  if (t.bt() == Type::BT_INT) {
+    t.typeId(0);
+  }
   auto it = _revmapmap.find(t.toInt());
   if (it != _revmapmap.end()) {
     return it->second;
