@@ -89,9 +89,10 @@ vector<string> MIPGurobiWrapper::getStdFlags() { return {"-i", "-p", "-s", "-v"}
 
 vector<string> gurobi_dlls() {
   const vector<string> versions = {
-      "954", "953", "952", "951",  // Potential future versions which should load correctly
-      "950", "912", "911", "910", "903", "902", "901", "900", "811", "810", "801",
-      "800", "752", "751", "750", "702", "701", "700", "652", "651", "650"};
+      "1013", "1012", "1011", "1010", "1003", "1002", "1001",  // Potential future versions which
+                                                               // should load correctly
+      "1000", "952", "951", "950", "912", "911", "910", "903", "902", "901", "900", "811", "810",
+      "801", "800", "752", "751", "750", "702", "701", "700", "652", "651", "650"};
   vector<string> dlls;
   string lastMajorVersion;
   for (const auto& version : versions) {
@@ -193,7 +194,8 @@ bool MIPGurobiWrapper::Options::processOption(int& i, std::vector<std::string>&
   std::string buf;
   if (cop.get("-i")) {
     flagIntermediate = true;
-  } else if (string(argv[i]) == "-f") {              // NOLINT: Allow repeated empty if
+  } else if (string(argv[i]) == "-f" ||
+             string(argv[i]) == "--free-search") {   // NOLINT: Allow repeated empty if
   } else if (string(argv[i]) == "--fixed-search") {  // NOLINT: Allow repeated empty if
     nFreeSearch = MIPGurobiWrapper::SearchType::FIXED_SEARCH;
   } else if (string(argv[i]) == "--uniform-search") {  // NOLINT: Allow repeated empty if
@@ -485,7 +487,7 @@ void MIPGurobiWrapper::doAddVars(size_t n, double* obj, double* lb, double* ub,
     }
   }
   _error = dll_GRBaddvars(_model, static_cast<int>(n), 0, nullptr, nullptr, nullptr, obj, lb, ub,
-                          &ctype[0], &pcNames[0]);
+                          ctype.data(), pcNames.data());
   wrapAssert(_error == 0, "Failed to declare variables.");
   _error = dll_GRBupdatemodel(_model);
   wrapAssert(_error == 0, "Failed to update model.");
@@ -913,7 +915,7 @@ void MIPGurobiWrapper::solve() {        // Move into ancestor?
   /// Solution callback
   output.nCols = static_cast<int>(colObj.size());
   _x.resize(output.nCols);
-  output.x = &_x[0];
+  output.x = _x.data();
   SolCallbackFn solcbfn = cbui.solcbfn;
   if (true) {  // NOLINT: Need for logging
     cbui.fVerb = fVerbose;
@@ -1015,7 +1017,7 @@ void MIPGurobiWrapper::solve() {        // Move into ancestor?
     assert(cur_numcols == colObj.size());
 
     _x.resize(cur_numcols);
-    output.x = &_x[0];
+    output.x = _x.data();
     _error = dll_GRBgetdblattrarray(_model, GRB_DBL_ATTR_X, 0, cur_numcols, (double*)output.x);
     wrapAssert(_error == 0, "Failed to get variable values.");
     if ((!_options->flagIntermediate || !cbui.printed) && (solcbfn != nullptr)) {
